#ifndef qsieve_header
#define qsieve_header


class CTinyIntVector
{
public:
  int akt_size; // aktuelle Gre des Vektors
  int max_size; // maximale Gre des Vektors
  int* v; // dynamisches Feld
  
  inline CTinyIntVector(const int MaxSize = 20) // Konstruktur
    {
      max_size=MaxSize;
      akt_size=0;
      v = new int [MaxSize];
    }
  inline CTinyIntVector(const int* Feld, const int Size)
    // Konstruktur: mit vorgegebenem Feld initialisieren
    {
      akt_size=max_size=Size;
      v = new int [Size];
      for(int i=0; i<Size; i++) v[i]=Feld[i];
    }
  inline ~CTinyIntVector() // Destruktor
    {
      delete [] v;
    }
  inline void reset()
    {
      akt_size=0;
    }
  inline void copy_from(const int* Feld, const int Size)
    {
      if (Size!=max_size) { delete [] v; v = new int [Size]; }
      akt_size=max_size=Size;
      for(int i=0; i<Size; i++) v[i]=Feld[i];
    }
  inline void resize(const int new_size)
    {
      if (new_size<akt_size)
	{ cerr << "Tiny-Vector: Resizing not possible!"; exit(1); }
      if (new_size==max_size) return; // nothing to do...
      int* vnew = new int [new_size];
      for (int i=0; i<akt_size; i++) vnew[i]=v[i];
      delete [] v; v=vnew; max_size=new_size;
    }
  inline void optisize() // Gre des dyn. Array optimieren
    {
      resize(akt_size);
    }
  inline void append(const int value)
    {
      if (akt_size>=max_size) 
	{
	  cerr << "TinyVector: Index " << akt_size
	       << " is out Of Range!"
	       << " Auto-Resizing by 5" << endl;
	  resize(akt_size+5);
	}
      if (!empty() && value<=v[akt_size-1])
	{ cerr << "TinyVector: unsortiertes append!" << endl; } 
      v[akt_size++]=value; 
    }
  inline void fast_append(const int value)
    {
      v[akt_size++]=value;
    }
  inline int size() const
    {
      return akt_size;
    }
  inline bool empty() const
    {
      return (akt_size==0);
    }
  inline int last() const
    {
      if (!empty()) return v[akt_size-1];
      else
	{
	  cerr << "TinyVector: no last Element!" << endl;
	  return -1;
	}
    }
};


const int no_factor = -1;
const int dynamic_factor = -2;
const int cached_dynamic_factor = -3;
const int special_factor = -4;

class CGleichungszeile
/* CGleichungszeile verwaltet eine Relation der Form
       Delta^2 = Produkt von Faktoren (mod N)
   Ziel ist es, die Menge der Faktoren durch Verknpfung von Relationen
   zu leeren. Das Produkt wird dann 1.
   Delta pat sich den Vernderungen dynamisch an.
*/ 
{
 public:
  int relevanter_Faktor;
 private:
  CTinyIntVector *Relation_sparse;
  BitString *Relation_dense;
  mpz_t Delta; // dynamischer Delta-Wert
 public:
  inline CGleichungszeile() /* Konstruktor */
    {
      relevanter_Faktor=no_factor;
      Relation_sparse = new CTinyIntVector;
      Relation_dense = NULL;
      mpz_init_set_ui(Delta,1);
    }
 public:
  inline ~CGleichungszeile() /* Destruktor */
    {
      if (Relation_sparse) { delete Relation_sparse; Relation_sparse=NULL; }
      if (Relation_dense) { delete Relation_dense; Relation_dense=NULL; }
      mpz_clear(Delta);
      relevanter_Faktor=no_factor;
    }
  
  CGleichungszeile(const signed int Siebpos, const short int HitCount=0); /* Konstruktur */
  /* "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */

 public:
  int groesster_Faktor_in_Relation() const
    /* (Index) des grter Faktor in der Relation */
    {
      if (Relation_sparse)
	if (Relation_sparse->empty()) return no_factor;
        else 
	  {
	    return Relation_sparse->last();
	  }
      else 
	{
	  return Relation_dense->last(1);
	  /*
	    int h = Relation_dense->last(1);
	    if (h>=0) return h;
	    else return no_factor;
	  */
	}
    }
  

  inline unsigned int SizeOfRelation() const
    {
      if (Relation_sparse) return Relation_sparse->size();
      else return Relation_dense->count(1);
    }
  
  unsigned int SizeOfRelation_verknuepft_mit(const CGleichungszeile& GL2) const;
  
  void Ausgabe() const; /* Gleichungszeile ausgeben */
  void convert_Relation_to_dense(); /* mache "dense" aus "sparse" */
  void convert_Relation_to_sparse(); /* mache "sparse" aus "dense" */
  inline bool Relation_ist_leer() /* gibt zurck, ob die Relation leer ist */
    const
    {
      return (relevanter_Faktor==no_factor);
      // if (Relation_sparse) return Relation_sparse->empty();
      // else return Relation_dense->last(1)==-1; 
    }	
  void kombinieren(const CGleichungszeile& GL2);
  /* die Vereinigungsmengen mit GL2 bilden und daraus die
     Schnittmengen entfernen. (-> also: symmetrische Differenz bilden)
     gibt zurck, ob die Relationsmenge leer ist, d.h. eine Kombination
     quadratischer Reste mit dem gleichen Wert gefunden wurde. */
  
  void Quadratische_Kongruenz_ausrechnen() const;
  void check_Relation(const Tfactor Restfaktor) const;
  void Kongruenzcheck() const; // test, ob Kongruenz noch in Ordnung 
 private:
  inline void swap(CGleichungszeile &GL2)
    { /* Inhalte der beiden Gleichungszeilen (referenzsemantisch) tauschen */
      CTinyIntVector *h1;
      BitString *h2;
      h1=Relation_sparse; Relation_sparse=GL2.Relation_sparse;
      GL2.Relation_sparse=h1;
      h2=Relation_dense; Relation_dense=GL2.Relation_dense;
      GL2.Relation_dense=h2;
      int h3;
      h3=relevanter_Faktor; relevanter_Faktor=GL2.relevanter_Faktor;
      GL2.relevanter_Faktor=h3;
      mpz_t h4;
      mpz_init_set(h4,Delta); mpz_set(Delta,GL2.Delta); mpz_set(GL2.Delta,h4);
      mpz_clear(h4);
    }
 public:
  CGleichungszeile operator= (const CGleichungszeile &GL)
    {
      cerr << "Zuweisung!!!! (nicht erlaubt)" << endl;
      exit(1);
      return GL;
    }
  bool operator< (const CGleichungszeile &GL2)
    const
    {	
      return relevanter_Faktor < GL2.relevanter_Faktor;
    }
  streampos speichern(ostream &out, const Tfactor factor=1, const short int HitCount=0) const;
  Tfactor kombinieren(istream &in, const streampos pos);
  Tfactor kombinieren(istream &in);
  friend ostream& operator<< (ostream& ostr, const CGleichungszeile& GL);
  friend void Relation_einfuegen(CGleichungszeile *GL);
  friend void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor);
  friend void bearbeite_daten_stream(iostream &daten);
  void Relation_optimieren();
};


ostream& operator<< (ostream& ostr, const CGleichungszeile& GL);
// Stream-Ausgabefunktion fr Gleichungszeile

void Relation_einfuegen(CGleichungszeile *GL);
// fgt Gleichungszeile in das Gleichungssystem ein

void Special_Factoren_durch_Primfaktor_splitten(const int Primfaktor);
// Specialfaktoren liegen auerhalb der Primzahlbasis.
// Mit ihnen wird nicht gesiebt. Sie sind auerdem zusammengesetzt.
// Wenn es gelingt, von einem Specialfaktor einen Primfaktor aus der dynamischen FB
// abzusplitten, dann bestehen gute Chancen, dadurch einen neuen dynamischen Faktor
// zu erhalten oder gar einen Treffer im Gleichungssystem, wenn der entstandene
// dynamische Faktor bereits in der dynamischen Faktorbasis bekannt ist.


typedef list<CGleichungszeile> TZeilenliste;


class CPolynom
{
  /* Implementierung der Polynomberechnungen.
     Variablennotation hlt sich (einigermaen) an die Beschreibung
     aus 
     "The Multiple Polynomial Quadratic Sieve,
     Robert D. Silverman,
     Mathematics Of Computation, Vol. 48, January 1987, p. 329-339
  */  
private:
  int M;
  mpz_t A,B,C,D,h0,h1,h2;
  mpz_t kN, kN_div2, D2_inv_mod_kN;
  mpz_t A2;
public:
  inline CPolynom() // Konstruktur
    {
      mpz_init(A); mpz_init(B); mpz_init(C); mpz_init(D);
      mpz_init(h0); mpz_init(h1); mpz_init(h2);
      mpz_init(kN); mpz_init(kN_div2); mpz_init(D2_inv_mod_kN);
      mpz_init(A2);
    }
  inline ~CPolynom() // Destruktor
    {
      mpz_clear(A); mpz_clear(B); mpz_clear(C); mpz_clear(D);
      mpz_clear(h0); mpz_clear(h1); mpz_clear(h2);
      mpz_clear(kN); mpz_clear(kN_div2); mpz_clear(D2_inv_mod_kN);
      mpz_clear(A2);
    }
  void erstes_Polynom_ermitteln(const mpz_t &fuer_kN, const int M);
  void naechstes_Polynom_ermitteln(const int step=0);
  void Werte_holen(const int Siebpos, mpz_t &Wurzelwert, mpz_t &Q) const;
  void testen(const signed int Siebpos=0);
  void Daten_sichern(ostream &ostr); // fr Recovery etc.
  void Daten_laden(istream &in); // fr Recovery etc.
  inline bool operator < (const mpz_t &K) const 
    {
      return mpz_cmp(D, K)<0;
    }
  friend void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset);
  friend void alle_Deltas_neu_berechnen();
};


void Deltas_berechnen_fuer(const TDynamicFactorRelation &DynRel, const int offset);
void alle_Deltas_neu_berechnen();
  
#endif
