#ifndef __UNIX_BUFFER__
#define __UNIX_BUFFER__

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <streambuf.h>
#include <iostream.h>

class unix_buffer : public streambuf {
protected:
  static const int buffersize = 1024;
  static const int pushbacksize = 10;
  int fd;
  bool is_open;
  char obuffer[buffersize];
  char ibuffer[buffersize+pushbacksize];

  int empty_buffer() {
    int count = pptr()-pbase();
    if (write(fd, obuffer, count) != count) {
      return EOF;
    }
    pbump(-count);
    return count;
  }

  virtual int overflow (int c) {
    if (c != EOF) {
      *pptr() = c;
      pbump(1);
    }
    if (empty_buffer() == EOF) {
      return EOF;
    }

    return c;
  }

  virtual int sync () {
    if (empty_buffer() == EOF) {
      return -1;
    }
    return 0;
  }

  virtual int underflow () {
    int count, pbcount;
    if (gptr() < egptr()) {
      return *gptr();
    }
    pbcount = gptr() - eback();
    if (pbcount > pushbacksize) {
      pbcount = pushbacksize;
    }
    if (pbcount > 0) {
      memcpy (ibuffer+(pushbacksize-pbcount), gptr()-pbcount, pbcount);
    }
    count = read(fd, ibuffer+pushbacksize, buffersize);
    if (count <= 0) {
      return EOF;
    }
    setg(ibuffer+(pushbacksize-pbcount), ibuffer+pushbacksize,
	 ibuffer+pushbacksize+count);
    return *gptr();
  }
  
public:
  unix_buffer(int _fd) {
    fd = _fd;
    is_open = true;
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }

  unix_buffer(char *host, int port) {
    struct sockaddr_in server;
    struct hostent *hp;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
      cerr << "Can't open socket" << endl;
      exit(1);
    }
    server.sin_family = AF_INET;
    hp = gethostbyname(host);
    if (hp == NULL) {
      cerr << "Unknown host: " << host << endl;
      exit(1);
    }
    memcpy((char*)&server.sin_addr, (char*)hp->h_addr, hp->h_length);
    server.sin_port = htons(port);
    if (connect(fd, (struct sockaddr *)&server, sizeof(server)) < 0) {
      cerr << "Can't connect socket" << endl;
      exit(1);
    }
    is_open = false;
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }

  virtual ~unix_buffer() {
    sync();
    if (!is_open) close(fd);
  }
};

class unix_io_client : public iostream {
protected:
  unix_buffer buf;
public:
  unix_io_client(char *host, int port) : iostream(&buf), buf(host, port) {
  }
};
  
int open_internet_port(int port, int qsize = 5)
{
  int sd;
  int val = 1;
  struct sockaddr_in server;

  if (port <= 0) return -1;

  sd = socket(AF_INET, SOCK_STREAM, 0);
  if (sd < 0) return sd;
  if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&val, sizeof(val)) <0)
    return -1;
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);
  if (bind(sd, (struct sockaddr *)&server, sizeof(server)) < 0) {
    return -1;
  }
  listen (sd, qsize);
  return sd;
}

#endif
